//////////////////////////////////////////////////////////////////////////
// Module	 : sx17mod.c												//
// Function  : Linux driver API for Sensoray model 617					//
// Copyright : (C) Sensoray 2003										//
//////////////////////////////////////////////////////////////////////////

/*
Copyright (C) 2003  Sensoray Co., Inc.  This program is free software;
you can redistribute it and/or modify it under the terms of the GNU
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

In addition, as a special exception, Sensoray Co., Inc., gives permission
to link the code of this program with the 'sx17core.o' library
and distribute linked combinations including the two. You must obey the
GNU General Public License in all respects for all of the code used other
than 'sx17core.o'. If you modify this file, you may extend this exception
to your version of the file, but you are not obligated to do so. If you
do not wish to do so, delete this exception statement from your version.
*/

//#define _RH_WMG_            //required to compile on WMG Linux

// The necessary header files

#include <linux/config.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#ifndef CONFIG_PCI
	#error "CONFIG_PCI undefined"
#endif

#include <linux/module.h>				// We're building a module.
#include <linux/kernel.h>				// We're doing kernel work.
#include <linux/fs.h>					// File operations support.
#include <linux/proc_fs.h>				// For create_proc_read_entry().
#include <linux/pci.h>					// Required for PCI.
#include <linux/poll.h>					// Poll declarations.
#include <linux/delay.h>				// Fast delay generators.
#include <asm/uaccess.h>				// For get_user() and put_user().
#include <asm/io.h>
#include <linux/highmem.h>
#include <linux/wrapper.h>
#include <asm/atomic.h>					// Atomic arithmetic
//#include <linux/time.h>

#include "sx17.h"						// IOCTL defines, etc.
#include "sx17mod.h"					// Declarations consistency check.
#include "sx17core.h"					// Interface to sx17core.o.

//Board index can be 0. request_irq doesn't like 0 as dev_id value
#define BOARD_TO_DEVID(x)				((x) + 1)
#define DEVID_TO_BOARD(x)				((x) - 1)

#define LOWORD(l)       ((WORD)(l))
#define HIWORD(l)		((WORD)((l)>>16))
#define LOBYTE(w)       ((BYTE)(w))
#define HIBYTE(w)		((BYTE)((w)>>8))
#define MAKELONG(l,h)	((ULONG)(h))<<16 | (l)

struct pci_dev	*bridge		= NULL;
struct pci_dev	*dev		= NULL;

void print_pci_config (void);
void process_irq (ULONG index);

///////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////  TYPES  ////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////  CONSTANTS  ////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

// Manifest constants ---------------------------------

#define TRUE				1
#define FALSE				0

#define INSTALL_PASS		0						// Module/board installation success code.
#define INSTALL_FAIL		-1						// Module/board installation fail code.

#define REGION_SIZE_617		512						// Size of SAA7146 application memory-mapped space (PCI region 0) occupied by the 617 board.

#define	BPP_MONO			1						//bytes per pixel, monochrome
#define	BPP_YUV				2						//bytes per pixel, YCrCb
#define	BPP_RGB				3						//bytes per pixel, RGB

#define BMPLINES			240						// bitmap max lines
#define	BMPPIXELS			640					    // bitmap max pixels per line
#define	BMPSIZE				BMPPIXELS * BMPLINES * BPP_RGB	// max bitmap buffer size (~460K)

#define DEBI_BLOCK			0x4000					// 16K
#define JPEG_BLOCKS			5
#define JPGSIZE				(JPEG_BLOCKS + 1) * DEBI_BLOCK  // jpeg buffer size (96K)
#define DEBUG_SIZE			0x4000
#define BUFSIZE				(JPGSIZE + BMPSIZE + 3 * PAGE_SIZE) * SYS_BUFS + PAGE_SIZE

#define BOARD_INSTALLED		0x00000001				//board detected and PCI registers mapped
#define BOARD_OPEN			0x00000002				//board opened, memory allocated

// Configuration constants ----------------------------

// Default values that are applied in place of missing insmod parameters:
#define DFLT_MAJOR			125						// Major device number.

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////  DATA STORAGE  ////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

static unsigned int			major					 		= DFLT_MAJOR;	// Major device number.
static int					num_boards;										// Board count obtained from pci scan.
static int					board_status	[MAX_BOARDS]	= { 0, 0, 0, 0};// BOARD_INSTALLED | BOARD_OPEN
static struct pci_dev		*s617_dev		[MAX_BOARDS];					// Pointer to pci_dev structures for model 617 boards.
static void					*log_base		[MAX_BOARDS];					// Logical base addresses of application registers.
static unsigned int			io				[MAX_BOARDS] 	= { 0, 0, 0, 0};// Physical base I/O address of application registers.
static SDATA617				sd617			[MAX_BOARDS];					// Data buffers
static wait_queue_head_t	wait_q			[MAX_BOARDS];					// Wait queues for processes that are waiting to read/write data.
static FIFO    				g_CmdFIFO		[MAX_BOARDS];					//command FIFO's
static FIFO    				g_BfrFIFO		[MAX_BOARDS];					//buffer FIFO's
static FIFO    				g_CapFIFO		[MAX_BOARDS];					//caption FIFO's
static int					stop_irq		[MAX_BOARDS];					//"stop interrupts" flag - set when quitting
static INT_DATA				intdata			[MAX_BOARDS];					//"static" data used by interrupt handler
static MODE					mode			[MAX_BOARDS];					//MODE settings
static int                  Q_DecNum        [MAX_BOARDS];                   //current decoder (for QueueCmd)
static int                  Q_PreviousCompfactor[MAX_BOARDS][4];            //                (for QueueCmd)
static char                 Q_PreviousCaption[MAX_BOARDS][4][65];
static int                  Q_PreviousCaptionmode [MAX_BOARDS][4];

////////////////////// D E B U G ///////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  FORWARD DECLARATION  /////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
static int 				AllocBuffers 		(int board);
void 					FIFOwrite 			(FIFO *pFIFO, FIFOELEMENT *pElement);
void 					FIFOread 			(FIFO *pFIFO, FIFOELEMENT *pElement);
int 					GetFreeBuffer 		(int board);
void 					FIFOdestroy 		(FIFO *pFIFO);
void 					fix845 				(void);

/////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////  INSMOD PARAMETERS //////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

// Declare the insmod parameters and establish linkage to their respective command line parameters.
//MODULE_PARM( major,		"h"									 );

///////////////////////////////////////////////////////////////////////////////////////
// Construct a list of pointers to the pci_dev structures for all model 617 boards.
// Populates s616_list[] and returns the count of model 616 boards.

#define SAA7146_VENDORID        0x1131		// Philips.
#define SAA7146_DEVICEID        0x7146		// SAA7146 PCI bridge.

#define S617_SUBVENDORID		0x6000		// Sensoray.
#define S617_SUBDEVICEID		0x0617		// Model 617.


static unsigned int find_617s( struct pci_dev **devlist )
{
	int				nboards		= 0;
	struct pci_dev	*dev		= NULL;
	unsigned short	sub_vendor;
	unsigned short	sub_device;

	// Repeat until we have found the maximum number of supported model 616 boards ...
	do {
		// Find next board that has the correct main ids.
		dev = pci_find_device( SAA7146_VENDORID, SAA7146_DEVICEID, dev );

		// Exit search loop if no more potential 617 boards are present.
		if ( !dev ) {
			break;
		}

		// Fetch the PCI sub-ids from the candidate board.
		pci_read_config_word( dev, PCI_SUBSYSTEM_VENDOR_ID, &sub_vendor );
		pci_read_config_word( dev, PCI_SUBSYSTEM_ID, &sub_device );

		// If sub-ids match, add board to the list.
		if ( ( sub_vendor == S617_SUBVENDORID ) && ( sub_device == S617_SUBDEVICEID ) ) {
			*devlist++ = dev;
			nboards++;
		}
	} while ( nboards < MAX_BOARDS );

	// Return the number of 617 boards that were found.
	return nboards;
}

/////////////////////////////////////////////////////////
// Memory allocation functions.

/* Given PGD from the address space's page table, return the kernel
 * virtual mapping of the physical memory mapped at ADR.
 */
static inline ULONG uvirt_to_kva (pgd_t *pgd, ULONG adr)
{
	ULONG 			ret = 0UL;
	pmd_t 			*pmd;
	pte_t 			*ptep, pte;

	if (!pgd_none (*pgd)) {
		pmd = pmd_offset (pgd, adr);
		if (!pmd_none (*pmd)) {
#ifdef _RH_WMG_
            ptep = pte_offset_map (pmd, adr);
#else
			ptep = pte_offset (pmd, adr);
#endif  //_RH_WMG_
			pte = *ptep;
			if (pte_present (pte)) {
				#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,23)
				ret = pte_page(pte);
				#else
				ret = (ULONG) page_address (pte_page (pte));
				#endif
				ret |= (adr & (PAGE_SIZE - 1));
			}
		}
	}
	return ret;
}

static inline ULONG kvirt_to_pa (ULONG adr)
{
	ULONG 			va, kva, ret;

	va = VMALLOC_VMADDR (adr);
	kva = uvirt_to_kva (pgd_offset_k (va), va);
//	printk("uv2kva(%08lX-->%08lX)\n", va, kva);
	ret = __pa(kva);
//	printk ("kv2pa(%08lX-->%08lX (bus %08lX)\n", kva, ret, virt_to_bus ((void *)kva));
	return ret;
}

static void * rvmalloc (ULONG size)
{
	void 			*mem;
	ULONG			adr, page;

	mem = vmalloc_32 (size);
	if (mem) {
		adr = (ULONG) mem;
//		printk ("vmalloc_32: %08lX\n", adr);
		while (size > 0) {
			page = kvirt_to_pa (adr);
			mem_map_reserve (virt_to_page (__va (page)));
//			printk ("mem_map_reserve: %08lX-->%08lX \n", page, (ULONG) __va (page));
			adr += PAGE_SIZE;
			size -= PAGE_SIZE;
		}
	} else {
		printk("vmalloc_32(%ld) failed\n", size);
	}
	return mem;
}

static void rvfree (void *mem, ULONG size)
{
	ULONG 			adr, page;

	if (mem) {
		adr = (ULONG) mem;
		while (size > 0) {
			page = kvirt_to_pa (adr);
			mem_map_unreserve (virt_to_page (__va (page)));
			adr += PAGE_SIZE;
			size -= PAGE_SIZE;
		}
		vfree (mem);
	}
}

/////////////////////////////////////////////////////////
// Release buffers used by a board.
static void FreeBuffers (int board)
{
	int i;

	for (i = 0; i < SYS_BUFS; i++) {
		if (sd617[board].sbuf[i].bmp.spt) {
			kfree ((void *)sd617[board].sbuf[i].bmp.spt);
			sd617[board].sbuf[i].bmp.spt = 0;
		}
		if (sd617[board].sbuf[i].bmp.srps) {
			kfree ((void *)sd617[board].sbuf[i].bmp.srps);
			sd617[board].sbuf[i].bmp.srps = 0;
		}
		if (sd617[board].sbuf[i].jpg.spt) {
			kfree ((void *)sd617[board].sbuf[i].jpg.spt);
			sd617[board].sbuf[i].jpg.spt = 0;
		}
		if (sd617[board].sbuf[i].jpg.srps) {
			kfree ((void *)sd617[board].sbuf[i].jpg.srps);
			sd617[board].sbuf[i].jpg.srps = 0;
		}
	}

	FIFOdestroy (&g_CmdFIFO[board]);
	FIFOdestroy (&g_BfrFIFO[board]);
	FIFOdestroy (&g_CapFIFO[board]);

	if (sd617[board].kbufstart) {
		rvfree (sd617[board].kbufstart, BUFSIZE);
		sd617[board].kbufstart = 0;
	}
}

///////////////////////////////////////////////////////////////////
// Interrupt processing function and handler, some helper functions

//Crosspoint constants
#define	CHAN_OFF				0xa				//channel off
int GetXP (int index, int chan)
{
	//returns current xpoint channel # for a given channel #
	int shifts;
	int xpchan;

	if (mode[index].boardver != BOARD_S4) {
		chan &= 3;
		shifts = (3 - chan) * 4;

		if (((LOWORD (intdata[index].xpdata) >> shifts) & 0xf) == CHAN_OFF) {
			xpchan = 9 + ((HIWORD (intdata[index].xpdata) >> shifts) & 0x7);
		} else {
			xpchan = 1 + ((LOWORD (intdata[index].xpdata) >> shifts) & 0x7);
		}
	} else {
		xpchan = chan + 1;
	}
	return xpchan;
}


void process_irq (ULONG index)
{
	ULONG				status;					//status buffer (used on mult occasions)
	ULONG				jsize;					//jpeg size
	int					cmd_ok, bfr;			//

	SX17CORE_dbg (index, 0);
	// 1: check last capture status, write output buffer
	////////////////////////////////////////////////////
	if (intdata[index].started) {                                      //if capture was started at all
		intdata[index].bfr.buffer.valid = 0;									//reset flags and size
		intdata[index].bfr.buffer.jpegsize = 0;
		intdata[index].bfr.buffer.incomplete = 0;
		intdata[index].bfr.buffer.zerr = 0;
		if (intdata[index].icmd.icommand.cmd.jpegon) {                        //if jpeg capture enabled
			status = SX17CORE_jpeg_status (index, &jsize);				//get jpeg capture status and data size
			if (status & CAPTURE_ST_INC) {								//"incomplete" error
				intdata[index].bfr.buffer.incomplete = 1;						//set flag
			} else if (status & CAPTURE_ST_ZERR) {						//compression error
				intdata[index].bfr.buffer.zerr = 1;
			} else {
				intdata[index].bfr.buffer.valid = 1;							//mark buffer as valid
				intdata[index].bfr.buffer.jpegsize = jsize;					//jpeg data size
			}
		} else if (intdata[index].icmd.icommand.cmd.bmpon) {					//if bitmap capture enabled
			status = SX17CORE_bmp_status (index);						//get bitmap capture status
			if (status) {
				intdata[index].bfr.buffer.incomplete = 1;						//set flag
			} else {
				intdata[index].bfr.buffer.valid = 1;							//mark buffer as valid
			}
		}
		//Copy data to temp BUFFER struct
		intdata[index].bfr.buffer.xpchan = GetXP (index, intdata[index].icmd.icommand.chan);
		intdata[index].bfr.buffer.reserved[0] = intdata[index].icmd.icommand.chan;
		intdata[index].bfr.buffer.vxp = intdata[index].vxp;
		intdata[index].bfr.buffer.videopresent = !(intdata[index].videostatus & VIDEO_NOT_PRESENT);
		intdata[index].bfr.buffer.dioin = intdata[index].dioin;
		intdata[index].bfr.buffer.compfactor = SX17CORE_get_compfactor (index);
		intdata[index].bfr.buffer.bright = intdata[index].decmode[intdata[index].icmd.icommand.chan].bright;
		intdata[index].bfr.buffer.contrast = intdata[index].decmode[intdata[index].icmd.icommand.chan].contrast;
		intdata[index].bfr.buffer.sat = intdata[index].decmode[intdata[index].icmd.icommand.chan].sat;
		intdata[index].bfr.buffer.hue = intdata[index].decmode[intdata[index].icmd.icommand.chan].hue;
		intdata[index].bfr.buffer.bi = intdata[index].bi;

        if (intdata[index].skip4 < 4) {
            intdata[index].bfr.buffer.valid = 0;
            ++intdata[index].skip4;
        }
		FIFOwrite (&g_BfrFIFO[index], (FIFOELEMENT *) &intdata[index].bfr);
		wake_up_interruptible (&wait_q[index]);
        intdata[index].old_icmd = intdata[index].icmd;
		intdata[index].started = FALSE;
	}

	// 2: get next command and a free buffer and start h/w
	///////////////////////////////////////////////////////
	cmd_ok = atomic_read (&g_CmdFIFO[index].contents);		//any commands in command FIFO?
	bfr = GetFreeBuffer (index);
	if ((bfr < SYS_BUFS) && cmd_ok) {						//command present AND buffer available
		sd617[index].sbuf[bfr].busy = 1;					//mark buffer as busy
		intdata[index].bi = bfr;							//store buffer index
		FIFOread (&g_CmdFIFO[index], &intdata[index].icmd);	//read next command from FIFO
		FIFOread (&g_CapFIFO[index], &intdata[index].caption);//read caption from captionFIFO

		//I2C operations (to current decoder)
		//I2C mux has to be set up before START, so that no
		//FPGA access is performed after DEBI xfers are enabled
		/////////////////////////////////////////
		SX17CORE_set_i2c_chan (index, intdata[index].icmd.icommand.chan);
		SX17CORE_read_decoder_status (index, intdata[index].icmd.icommand.chan, &intdata[index].videostatus);
		//Set to previous decoder now, so that no FPGA accesses are done after START
		SX17CORE_set_i2c_chan (index, intdata[index].old_icmd.icommand.chan);

		//START sequence
		////////////////////
		SX17CORE_start (index, &intdata[index].icmd.icommand, &intdata[index].caption.caption[0]);	//start compression chip
		if (mode[index].boardver != BOARD_S4) {
			SX17CORE_setXPs (index, &intdata[index]);
			//Set video out switch and store value
			intdata[index].vxp = SX17CORE_set_vout (index, intdata[index].icmd.icommand.cmd.vxp, intdata[index].vxp);
		}

		//If any kind of capture enabled - start RPS
		if (intdata[index].icmd.icommand.cmd.jpegon || intdata[index].icmd.icommand.cmd.bmpon) {
			ULONG jpgrps = intdata[index].icmd.icommand.cmd.jpegon ? sd617[index].sbuf[bfr].jpg.prps : 0;
			ULONG bmprps = intdata[index].icmd.icommand.cmd.bmpon ? sd617[index].sbuf[bfr].bmp.prps : 0;
			SX17CORE_StartRPS (index, jpgrps, bmprps);
			intdata[index].started = TRUE;
		}

		//I2C operations (to previous decoder and DIO)
		SX17CORE_i2c_ops (index, &intdata[index]);

		//Save b/c/s/h parameters to decmode (to be copied to BUFFER from there)
		intdata[index].decmode[intdata[index].old_icmd.icommand.chan].bright = intdata[index].old_icmd.icommand.cmd.bright;
		intdata[index].decmode[intdata[index].old_icmd.icommand.chan].contrast = intdata[index].old_icmd.icommand.cmd.contrast;
		intdata[index].decmode[intdata[index].old_icmd.icommand.chan].sat = intdata[index].old_icmd.icommand.cmd.sat;
		intdata[index].decmode[intdata[index].old_icmd.icommand.chan].hue = intdata[index].old_icmd.icommand.cmd.hue;
	}

	SX17CORE_dbg (index, 1);
}

DECLARE_TASKLET(process_irq0, process_irq, 0);
DECLARE_TASKLET(process_irq1, process_irq, 1);

static void irq_handler (int irq, void *dev, struct pt_regs *regs)
{
	int 				board = DEVID_TO_BOARD((int) dev);
	ULONG				irq_status = 0;

	irq_status = SX17CORE_irq_check (board);
	if (!irq_status)					//not our board
		return;
	SX17CORE_irq_disable (board);		//disable board's irq
	if (stop_irq[board])				//if STOP flag set - return with interrupts disabled
		return;

    switch (board) {
	    case 0:
	        tasklet_schedule (&process_irq0);
			break;
	    case 1:
	        tasklet_schedule (&process_irq1);
			break;
		default:
		    break;
	}

//    process_irq (board);

	SX17CORE_irq_enable (board);		//enable board's irq
	return;
}

static int init_board( int board )
{
	ECODE				ecode;
    int                 i;

	//Initialize hardware
	if ((ecode = SX17CORE_board_init (board, (unsigned long)log_base[board]))) {	// Attempt to init the driver core.  If init failed ...
		printk( "x17 driver: x17(%d) did not initialize, error %d.\n", board, ecode);
		return INSTALL_FAIL;
	}
    memset (&intdata[board], 0, sizeof(INT_DATA));
	intdata[board].xpdata = XP_INIT;
	intdata[board].vxp = VXP_INIT;
    intdata[board].skip4 = 0;
    Q_DecNum[board] = 0;
    for (i = 0; i < 4; i++) {
        Q_PreviousCompfactor[board][i] = 0x100;     //some value for the very first frame
    }
    //initialize caption data
    for (i = 0; i < 4; i++) {
        Q_PreviousCaptionmode[board][i] = 0;
    }

	//Allocate buffers
	if (AllocBuffers (board)) {					// returns 0 if success
		printk ("x17 driver: buffer alloc failed for board %d\n", board);
		return INSTALL_FAIL;
	}

	//Install interrupt
	if (request_irq (s617_dev[board]->irq, irq_handler, SA_SHIRQ, "sx17", (void *)BOARD_TO_DEVID(board))) {
		printk ("x17 driver: interrupt installation failed for board %d\n", board);
		return INSTALL_FAIL;
	}
	stop_irq[board] = 0;

	// Indicate success/fail.
	return INSTALL_PASS;
}

/////////////////////////////////////////////////////////
// Release resources used by a board.
static void release_resources ( int i )
{
	iounmap(log_base[i]);
	release_mem_region( io[i], REGION_SIZE_617 );
}

/////////////////////////////////////////////////////////
// Close a board.
static void board_close (int i)
{
	if (board_status[i] & BOARD_OPEN) {
		stop_irq[i] = 1;					//set the flag to disable interrupts
		SX17CORE_board_close (i);			//stop the board
		free_irq (s617_dev[i]->irq, (void *) BOARD_TO_DEVID(i));
	}
	FreeBuffers (i);						//Release all memory; safe to call even if not alloc'd
}

/////////////////////////////////////////////////////////////
// Reset all installed boards and terminate the driver.
static void remove_driver( void )
{
	int i;

	for ( i = 0; i < num_boards; i++ ) {
		if (board_status[i] & BOARD_INSTALLED) {			// If the board is installed
			board_close (i);
			release_resources( i );
		}
	}

	// Terminate the driver.
	unregister_chrdev(major, "sx17");
}

//	FIFO functions

ECODE FIFOcreate (FIFO *pFIFO, DWORD size)
{
	//pFIFO - pointer to FIFO structure;
	//size - FIFO size (elements);
	ECODE ecode = 0;

	if ((pFIFO->start = (FIFOELEMENT *) kmalloc (size * sizeof(FIFOELEMENT), GFP_KERNEL)) == NULL) {
		ecode = ERR_MEM_ALLOC;
	} else {
		pFIFO->read = pFIFO->start;
		pFIFO->write = pFIFO->start;
		pFIFO->end = pFIFO->start + (size - 1);
		pFIFO->size = size;
		atomic_set (&pFIFO->contents, 0);
	}
	return ecode;
}

void FIFOdestroy (FIFO *pFIFO)
{
	//pFIFO - pointer to FIFO structure;
	if (pFIFO->start)
        kfree (pFIFO->start);
	pFIFO->start = NULL;
	return;
}


void FIFOread (FIFO *pFIFO, FIFOELEMENT *pElement)
{
//	FIFO.contents MUST be checked to be > 0 prior to calling FIFOread
	FIFOELEMENT *pfe;

	pfe = pFIFO->read;
	if (pFIFO->read == pFIFO->end) {
		pFIFO->read = pFIFO->start;			//roll over
	} else {
		++pFIFO->read;
	}
	atomic_dec (&pFIFO->contents);
	*pElement = *pfe;
}

void FIFOwrite (FIFO *pFIFO, FIFOELEMENT *pElement)
{
//	If FIFO.contents >= FIFO.size FIFOwrite does nothing
	if (atomic_read (&pFIFO->contents) < pFIFO->size) {
		*pFIFO->write = *pElement;
		if (pFIFO->write == pFIFO->end) {
			pFIFO->write = pFIFO->start;			//roll over
		} else {
			++pFIFO->write;
		}
		atomic_inc (&pFIFO->contents);
	}
	return;
}

int GetFreeBuffer (int index)
{
	int i;
	for (i = 0; i < SYS_BUFS; i++) {
		if (sd617[index].sbuf[i].busy == 0) {
			break;
		}
	}
	return i;
}


static int setmode (int board, MODE *pmode)
{
	ECODE			ecode;

	//clear fifo's
	atomic_set (&g_CmdFIFO[board].contents, 0);
	atomic_set (&g_BfrFIFO[board].contents, 0);
	atomic_set (&g_CapFIFO[board].contents, 0);

	SX17CORE_set_bufrs (pmode, &sd617[board]);

	ecode = SX17CORE_setmode (board, pmode);
	mode[board] = *pmode;
	return ecode;
}

void QueueCmd (int board, COMMAND *cmd)
{
    ICOMMAND    icmd;                               //buffer for "internal" command
    int         buffer;
    char        CaptionBuf[65];
    char        Caption[65];

	if (atomic_read (&g_CmdFIFO[board].contents) < SYS_BUFS) {		//check if any room in FIFO
        //set decoder number
        if (mode[board].boardver == BOARD_S4) {
            if ((cmd->xpchan < 1) || (cmd->xpchan > 4)) {
                icmd.chan = 0;
            } else {
                icmd.chan = cmd->xpchan - 1;
            }
        } else {
            icmd.chan = Q_DecNum[board];                         //assign to next decoder
        }
        icmd.cmd = *cmd;                                       //copy COMMAND struct (pointer will become obsolete)

        buffer = icmd.cmd.compfactor;                          //swap compfactor and PreviousCompfactor
        icmd.cmd.compfactor = Q_PreviousCompfactor[board][Q_DecNum[board]];//..
        Q_PreviousCompfactor[board][Q_DecNum[board]] = buffer;             //..

        //Need to do the same for caption and caption on/off flag
        strncpy (CaptionBuf, cmd->pCaption, 65);
        buffer = cmd->captionmode;
        strncpy (Caption, Q_PreviousCaption[board][Q_DecNum[board]], 65);
        icmd.cmd.captionmode = Q_PreviousCaptionmode [board][Q_DecNum[board]];
        strncpy (Q_PreviousCaption[board][Q_DecNum[board]], CaptionBuf, 65);
        Q_PreviousCaptionmode[board][Q_DecNum[board]] = buffer;

        FIFOwrite (&g_CapFIFO[board], (FIFOELEMENT *) Caption);
        FIFOwrite (&g_CmdFIFO[board], (FIFOELEMENT *) &icmd);
		if (Q_DecNum[board] < 3) {
            ++Q_DecNum[board];
        } else {
            Q_DecNum[board] = 0;
        }
    }
    return;
}



//////////////////////////////////////////////////////////////////////////////////////////////
// File access functions
static int sx17_open( struct inode *inode, struct file *file )
{
	int board = MINOR( inode->i_rdev );

	// Check validity of minor (board) number.
	if ((unsigned int) board > num_boards) {
		printk ("Error: invalid minor device number\n");
		return -ENODEV;
	}

	if ((board_status[board] & BOARD_INSTALLED) == 0) {
		printk ("Open failed for x17(%d): board not attached\n", board);
		return -ENODEV;
	}

	if (board_status[board] & BOARD_OPEN) {
		printk ("Open failed for x17(%d): already open\n", board);
		return -ENODEV;
	}

	// Init the board and take it to its default state.
	if (init_board (board) == INSTALL_FAIL ) {
		printk ("Open failed for x17(%d): initialization error\n", board);
		return -ENODEV;
	} else {
		board_status[board] |= BOARD_OPEN;				// Indicate board has been successfully open.
	}

	init_waitqueue_head (&wait_q[board]);
	// One more app is using this module.
	MOD_INC_USE_COUNT;
	printk ("x17 driver: sx17(%d) opened successfully\n", board);	
	return 0;
}

////////////////////////////////////////////////////////////////////////////
// CLOSE METHOD - release the device and decrement the usage counters.

static int sx17_release( struct inode *inode, struct file *file )
{
	int board = MINOR( inode->i_rdev );

	board_close (board);
	board_status[board] &= ~BOARD_OPEN;
	// One less app is using this module.
	MOD_DEC_USE_COUNT;
	printk ("x17 driver: sx17(%d) closed successfully\n", board);

	return 0;
}

/////////////////////////////////////////////////////////////////////////////
// MMAP method: allocates buffer memory
// and prepares it for mapping into user space

static int sx17_mmap (struct file *file, struct vm_area_struct *vma)
{
	//Note: sx17 map method does NOT support offsets into mapped memory
	ULONG size = vma->vm_end - vma->vm_start;
	ULONG start = vma->vm_start;
    ULONG page, pos;
	int board = MINOR( file->f_dentry->d_inode->i_rdev );

	vma->vm_flags |= VM_LOCKED;


//	printk ("sx17_mmap: board %d, adr %08lX, size %08lX\n", board, start, size);
	if (size > BUFSIZE - PAGE_SIZE) {
		printk ("sx17_mmap: size %ld > max %ld\n", size, BUFSIZE - PAGE_SIZE);
		return -EINVAL;
	}

	if (!(board_status[board] & BOARD_INSTALLED)) {
		printk ("sx17_mmap: board not installed\n");
		return -ENODEV;
	}

	if (!sd617[board].kbuf) {
		printk ("sx17_mmap: buffer not allocated\n");
		return -EINVAL;
	}

	pos = (ULONG) sd617[board].kbuf;
	while (size > 0) {
		page = kvirt_to_pa (pos);
#ifdef _RH_WMG_
		if (remap_page_range (vma, start, page, PAGE_SIZE, PAGE_SHARED))
#else
		if (remap_page_range (start, page, PAGE_SIZE, PAGE_SHARED))
#endif  //_RH_WMG_
			return -EAGAIN;
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
    return 0;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////
// ioctl method - the 'misc.' function of the driver.
static int sx17_ioctl( struct inode *inode, struct file *file, unsigned int iocmd, unsigned long val )
{
	int 			board = MINOR( inode->i_rdev );		// Board number.
	int 			retval = 0;							// Init return value in case command does not produce one.
	COMMAND 		cmd;								// buffer for COMMAND struct
	BUFFER			bfr;								// buffer for BUFFER struct

	// Decode the ioctl command code.
	switch (iocmd) {
		case SX17_IOC_GETMAPSIZE:
			retval = BUFSIZE - PAGE_SIZE;
			break;

		case SX17_IOC_SETMODE:
			retval = setmode (board, (MODE *) val);
			SX17CORE_irq_enable (board);			//enable board's irq
			break;

		case SX17_IOC_GETOFFSETS:
			copy_to_user ((void *) val, (void *) (sd617[board].oset), sizeof(OFFSET) * SYS_BUFS);
			break;

		case SX17_IOC_QUEUECMD:
			copy_from_user ((void *)&cmd, (void *) val, sizeof(COMMAND));
			QueueCmd (board, &cmd);
			break;

		case SX17_IOC_GETBUFR:
			retval = atomic_read (&g_BfrFIFO[board].contents);
			if (retval) {				//return value (non-zero if data returned)
				FIFOread (&g_BfrFIFO[board], (FIFOELEMENT *) &bfr);
				copy_to_user ((void *)val, (void *)&bfr, sizeof (BUFFER));
			}
			break;
			
		case SX17_IOC_RELEASEBUFR:
			sd617[board].sbuf[(int) val].busy = 0;
			break;

		default:
			retval = -EINVAL;
			break;	// Unsupported IOCTL command.
	}
	return retval;
}

////////////////////////////////////////////////////////////////////////////
// POLL METHOD - polls the device for readiness.
unsigned int sx17_poll( struct file *file, poll_table *wait )
{
	// Get the board handle.
	int board = MINOR( file->f_dentry->d_inode->i_rdev );

	// Add this board's list of blocked processes to the poll queue.
	poll_wait( file, &wait_q[board], wait );

	// Return appropriate flag if buffer FIFO contains data.
	if (atomic_read (&g_BfrFIFO[board].contents))
		return POLLIN;					// Readable

	// Indicate Not Ready.
	return 0;
}


//////////////////////////////////////////////////////////////////////////////////////////////
// The access methods are placed in a table of valid file operations.  This driver supports
// open, release (close), read, write and ioctl. Use the new gcc extension syntax.

struct file_operations sx17_fops = {
/*	read:		sx17_read,
	write:		sx17_write,*/
//    owner:      THIS_MODULE,
	poll:		sx17_poll,
	ioctl:		sx17_ioctl,
	release:	sx17_release,
	open:		sx17_open,
	mmap:		sx17_mmap
};

///////////////////////////////////////////////////////////////////////////////////
// Allocate resources used by a board.  Returns INSTALL_PASS or INSTALL_FAIL.
// This is called when the driver is installed. It does not allocate any
// resources other than PCI memory for board's register access.
// Everything else is allocated during "open"
static int attach_board ( int i )
{
	// Get the physical base address of the SAA7146 application register set.
	// These registers are accessed through PCI region 0.
	io[i] = pci_resource_start( s617_dev[i], 0 );

	// Abort if the requested memory space is not available.
	if ( check_mem_region( io[i], REGION_SIZE_617 ) ) {
		printk( "\nx17(%d) address space at 0x%8.8X is already attached.\n", i, io[i] );
		return INSTALL_FAIL;
	}

	// Open access to board's memory space.
	request_mem_region( io[i], REGION_SIZE_617, "sx17" );

	// Get the logical base memory address to be used for accessing the SAA7146 registers.
	log_base[i] = ioremap( io[i], REGION_SIZE_617 );
//	printk ("IRQ: %d\n", s617_dev[i]->irq);

	// Indicate success.
	return INSTALL_PASS;
}

///////////////////////////////////////////////////////////////////////////////////
// Buffer allocation functions.
static int AllocBuffers (int board)
{
	int 		i;
	ULONG		ptr, bsize;
	ULONG		current_offset, current_addr, current_size, *current_pt_addr;
	ECODE		ecode;

	memset ((void *) &sd617[board], 0, sizeof(SDATA617));

	//Allocate page table and RPS buffers first
	for (i = 0; i < SYS_BUFS; i++) {
		//Bitmap page table
		ptr = (ULONG) kmalloc (2 * PAGE_SIZE, GFP_KERNEL | __GFP_DMA);
		if (ptr == 0)
			return INSTALL_FAIL;
		sd617[board].sbuf[i].bmp.spt = ptr;
		//adjust to the start of next page, in case starts in the middle
		sd617[board].sbuf[i].bmp.lpt = (ptr + PAGE_SIZE) & ~(PAGE_SIZE - 1);
		sd617[board].sbuf[i].bmp.ppt = virt_to_bus ((void *) sd617[board].sbuf[i].bmp.lpt);

		//Bitmap RPS buffer
		ptr = (ULONG) kmalloc (2 * PAGE_SIZE, GFP_KERNEL | __GFP_DMA);
		if (ptr == 0)
			return INSTALL_FAIL;
		sd617[board].sbuf[i].bmp.srps = ptr;
		//adjust to the start of next page
		sd617[board].sbuf[i].bmp.lrps = (ptr + PAGE_SIZE) & ~(PAGE_SIZE - 1);
		sd617[board].sbuf[i].bmp.prps = virt_to_bus ((void *) sd617[board].sbuf[i].bmp.lrps);

		//JPEG page table
		ptr = (ULONG) kmalloc (2 * PAGE_SIZE, GFP_KERNEL | __GFP_DMA);
		if (ptr == 0)
			return INSTALL_FAIL;
		sd617[board].sbuf[i].jpg.spt = ptr;
		//adjust to the start of next page
		sd617[board].sbuf[i].jpg.lpt = (ptr + PAGE_SIZE) & ~(PAGE_SIZE - 1);
		sd617[board].sbuf[i].jpg.ppt = virt_to_bus ((void *) sd617[board].sbuf[i].jpg.lpt);

		//JPEG RPS buffer
		ptr = (ULONG) kmalloc (2 * PAGE_SIZE, GFP_KERNEL | __GFP_DMA);
		if (ptr == 0)
			return INSTALL_FAIL;
		sd617[board].sbuf[i].jpg.srps = ptr;
		//adjust to the start of next page
		sd617[board].sbuf[i].jpg.lrps = (ptr + PAGE_SIZE) & ~(PAGE_SIZE - 1);
		sd617[board].sbuf[i].jpg.prps = virt_to_bus ((void *) sd617[board].sbuf[i].jpg.lrps);
	}
	
	//Allocate FIFO's
	ecode = FIFOcreate (&g_CmdFIFO[board], SYS_BUFS);
	if (ecode)
		return INSTALL_FAIL;
	ecode = FIFOcreate (&g_BfrFIFO[board], SYS_BUFS);
	if (ecode)
		return INSTALL_FAIL;
	ecode = FIFOcreate (&g_CapFIFO[board], SYS_BUFS);
	if (ecode)
		return INSTALL_FAIL;

	//Allocate mmap buffer
	bsize = BUFSIZE;
//	printk ("Allocating %ld K\n", bsize / 1024);
	if (!(sd617[board].kbufstart = rvmalloc (bsize)))
		return INSTALL_FAIL;
	//Adjust to page start
	ptr = (ULONG) sd617[board].kbufstart;
	ptr += PAGE_SIZE;
	ptr &= ~(PAGE_SIZE - 1);
	sd617[board].kbuf = (void *) ptr;

	//Do page mapping for all components of mmap buffer:
	//	store phys addr of each page in corresponding page table,
	//	store buffer offsets
	current_offset = 0;
	current_addr = (ULONG) sd617[board].kbuf;
	for (i = 0; i < SYS_BUFS; i++) {
		//Bitmap
		current_pt_addr = (ULONG *) sd617[board].sbuf[i].bmp.lpt;		//current logical addr of page table
		sd617[board].oset[i].obmp = current_offset;						//store buffer offset
		for (current_size = 0;
			 current_size < BMPSIZE;
			 current_size += PAGE_SIZE, current_addr += PAGE_SIZE, current_offset += PAGE_SIZE) {
			*current_pt_addr++ = kvirt_to_pa (current_addr);			//store phys addr of current page in page table
		}
		//Jpeg
		current_pt_addr = (ULONG *) sd617[board].sbuf[i].jpg.lpt;
		sd617[board].oset[i].ojpg = current_offset;
		for (current_size = 0;
			 current_size < JPGSIZE;
			 current_size += PAGE_SIZE, current_addr += PAGE_SIZE, current_offset += PAGE_SIZE) {
			*current_pt_addr++ = kvirt_to_pa (current_addr);
		}
	}

	return INSTALL_PASS;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////////
// Install the driver and all declared boards.
static int install_driver( void )
{
	int 			i;
	int				attached, rtnval;

	// Assume we are installing 616s: scan the PCI bus and construct a list of all model 616 boards.
	num_boards	= find_617s (s617_dev);
	printk( "x17 driver: %d board(s) detected.\n", num_boards);

	// Terminate driver if no boards are available.
	if ( num_boards == 0 ) {
		return INSTALL_FAIL;
	}
	
	//Fix Intel's 845 and newer chipsets' PCI latency issue
	fix845 ();

	// Attempt to register the device with the kernel, and if unsuccessful, report the problem and terminate the driver.
	if (register_chrdev (major, "sx17", &sx17_fops)) {
		printk( "x17 driver: unable to get major number %i.\n", major );
		return INSTALL_FAIL;
	}

	// Attach all declared boards:
	for ( i = 0, attached = 0; i < num_boards; i++ ) {
		rtnval = attach_board (i);
		if (rtnval == INSTALL_PASS) {
			board_status[i] |= BOARD_INSTALLED;
			++attached;
		}
	}
/*	// Make public the function that supplies "proc" info to the system.
	create_proc_read_entry( "sx16",            // file name
	                  	    0,                 // default mode
                       	    NULL,              // parent dir
                       	    sx16_read_procmem, // file generation function
                       	    NULL);             // client data
*/
	return attached ? INSTALL_PASS : INSTALL_FAIL; // success if at least 1 board attached
}

////////////////////////////////////////////////////////////////////////////
// INSTALL METHOD.  This is called when the module is loaded.
// Register the module and install all declared boards.

int init_module( void )
{
	int rtnval;

	printk( "\nInstalling x17 driver...\n" );

	printk( ( ( rtnval = install_driver() ) == INSTALL_PASS ) ? "x17 driver successfully installed!\n\n" : "x17 driver installation failed!\n\n" );
	return rtnval;
}

////////////////////////////////////////////////////////////////////////////
// REMOVE METHOD.  This is called when the module is unloaded.
// Reset and release each installed board, terminate driver.

void cleanup_module( void )
{
	remove_driver();
	printk( "x17 driver removed!\n" );
}




//////////////////////////////////////////////////////////////////////////////////////////////////////
// ioctl method - the 'misc.' function of the driver.




//////////////////////////////////////////////////////////////////////////////////////////////////////
// misc helper functions of the driver.
void SX17MOD_xdelay (int us)
{
	udelay (10 * us);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
// The following function is necessary to provide adequate PCI performance for the 617 on some
// motherboards built with Intel's chipsets (845 and newer). Presumably because of an error in
// Intel's documentation most motherboards don't provide means of modifying PCI latency setting
// through BIOS, which causes a lot of 617 data transfer errors.
//////////////////////////////////////////////////////////////////////////////////////////////////////

#define INTEL_VID       0x8086              //Vendor ID for all Intel devices
#define PCIBRIDGE_DID   0x244e              //Device ID for PCI bridge

//IDE controller in FB (ICH6)
#define IDE_DID_ICH6    0x266f              //Device ID of IDE controller in ICH6
#define SATA_DID_ICH6   0x2651              //Device ID of SATA controller in ICH6
#define SATA_DID_ICH6W  0x2652              //Device ID of SATA controller in ICH6W

//Offsets and values to be fixed (VALUE is OR'ed with register contents)
#define REVIDADDR       0x8
#define OFFSET_FB       0x4c
#define OFFSET_DB       0x50
#define VALUE_FB        0x20
#define VALUE_DB        0x04


void fix845 (void)
{
	u32           	res = 0, val = 0;                //DWORD buffer
    int             bridgeID    = 0;                //bridge ID index
    char            *suff[2] = {"BA/CB/DB/EB", "FB/FR/GR"};

    bridge = pci_find_device( INTEL_VID, PCIBRIDGE_DID, bridge );   //look for PCI bridge that has to be fixed
    if (bridge) {                                                   //if found
        printk ("Intel 82801 bridge detected: ");
        //Find out which ICH model
        //RevID's:
        //BA(ICH2): 0x01-0x0c
        //DB(ICH4): 0x81-0x82
        //EB(ICH5): 0xc2
        //FB/FR(ICH6): 0xd3, 0xd4, 0xd5
        //GR(ICH7): 0xe1
        pci_read_config_dword (bridge, REVIDADDR, &res);
        if ((res & 0xf0) >= 0xc0) {
            bridgeID = 1;           //FB/FR/GR and above
        }
        printk ("%s\n", suff[bridgeID]);

        //Apply the fix
        switch (bridgeID) {
            case 0:                 //BA/CB/DB/EB
                pci_read_config_dword (bridge, OFFSET_DB, &res);
                printk ("Before fix 0x%08X\n", res);
                res |= VALUE_DB;
                pci_write_config_dword (bridge, OFFSET_DB, res);
                //read again
                pci_read_config_dword (bridge, OFFSET_DB, &val);
                printk ("                   after fix: 0x%08X\n", res);
                break;

            case 1:                 //FB/FR/GR
                pci_read_config_dword (bridge, OFFSET_FB, &res);
                printk ("Before fix 0x%08X\n", res);
                res |= VALUE_FB;
                pci_write_config_dword (bridge, OFFSET_FB, res);
                //read again
                pci_read_config_dword (bridge, OFFSET_FB, &val);
                printk ("          after fix: 0x%08X\n", val);
                break;

            default:
                printk ("Unrecognized bridge\n");
                break;
        }
    }
}

void SX17MOD_TranslateCaption (int index, char *pCaptionOut, char *pCaptionIn, int chan, int offset)
{
	int                 rows, cols, i, xpchan;
	int                 charswrtn = 0, /*msdigits,*/ leftinline;
	BOOL                zerochar = FALSE;
/*	char                timestr[40];
	char                str[16];
    struct timeval      tv;
    struct tm           *ptm;
    long                ms;*/


    xpchan = GetXP (index, chan);
    xpchan += offset;
	if (xpchan > 99) xpchan = 99;
	if (xpchan < 0) xpchan = 0;

	switch (mode[index].captionformat) {
		case CAP8X8:
			rows = 8;
			cols = 8;
			break;
		case CAP16X4:
			rows = 4;
			cols = 16;
			break;
		case CAP32X2:
			rows = 2;
			cols = 32;
			break;
		default:
			rows = 4;
			cols = 16;
			break;
	}


	while (charswrtn < CAPTION_LENGTH) {
		if (!zerochar) {
			if (*pCaptionIn == 0) {
				zerochar = TRUE;
			} else if (*pCaptionIn == '^') {		//special char: command
				++pCaptionIn;
				if (*pCaptionIn == 'n') {			//	command: new line
					++pCaptionIn;
					leftinline = cols - charswrtn % cols;
					for (i = 0; i < leftinline; i++) {
						*pCaptionOut++ = ' ';		//fill the rest of the line w/spaces
						++charswrtn;
					}
				} else if (*pCaptionIn == 'c') {	//  command: channel#
					++pCaptionIn;
					if (xpchan >= 10) {
						if (charswrtn < CAPTION_LENGTH) {
							*pCaptionOut++ = 0x30 + xpchan / 10;
						}
					} else {
						if (charswrtn < CAPTION_LENGTH) {
							*pCaptionOut++ = 0x30;
						}
					}
					++charswrtn;
					if (charswrtn < CAPTION_LENGTH) {
						*pCaptionOut++ = 0x30 + xpchan % 10;
					}
					++charswrtn;
				} else {							//special char not followed by a legal char
					++pCaptionIn;
					*pCaptionOut++ = '^';
					++charswrtn;
				}
			} else {								//normal char
				*pCaptionOut++ = *pCaptionIn++;
				++charswrtn;
			}
		} else {									//if zero detected
			*pCaptionOut++ = ' ';					//fill w/spaces
			++charswrtn;
		}
	}
}


void print_pci_config (void)
{
    int i;
    u32 res;

    if (bridge) {
        for (i = 0x4c; i < 0x50; i += 4) {
            pci_read_config_dword (bridge, i, &res);
            printk ("%04X    %08X\n", i, res);
        }
    } else {
        printk ("No bridge\n");
    }
}


// Set driver license.
MODULE_LICENSE( "GPL" );


